
#include <mcx.h>

          EXPORT mandelbrot, SHIFT
          CONST WORKSPACESIZE=100
          CONST SHIFT=13

INIT
KILL
  ;on entry
  ;first word on stack is workspace pointer
          MOV       pc,r14


mandelbrot
  ;on entry
  ;    r0 -> screen definition
  ;          +0     start address
  ;          +4     size x
  ;          +8     size y
  ;    r1 -> area to calculate
  ;          +0     minx
  ;          +4     miny
  ;          +8     stepx
  ;          +12    stepy
  ;       coords are in 19.13 fixed point
  ;first word on stack is workspace pointer
          STMFD     r13!,{r0-r12,r14}
          MOV       r10, r0
          MOV       r11, r1
          LDR       r9,[r10,#0]         ;screen
          LDR       r8,[r10,#8]         ;size y
          LDR       r2,[r11,#4]         ;min y
          LDR       r3,[r11,#8]         ;step x
          LDR       r4,[r11,#12]        ;step y
loop_y
           LDR      r1,[r11,#0]         ;min x
           LDR      r6,[r10,#4]         ;size x
loop_x
            MOV     r0,r5
            BL      do_one_pixel
            AND     r0,r0,#15

            LDRB    r5,[r9]
            TST     r6,#1
            ANDEQ   r5,r5,#15
            ORREQ   r5,r5,r0
            ANDNE   r5,r5,#15<<4
            ORRNE   r5,r5,r0,LSL#4
            STRB    r5,[r9]
            ADDNE   r9,r9,#1

           ADD      r1,r1,r3            ;next x
           SUBS     r6,r6,#1
           BGT      loop_x
          
          ADD       r2,r2,r4            ;next y
          SUBS      r8,r8,#1
          BGT       loop_y
          LDMFD     r13!,{r0-r12,pc}
          
do_one_pixel
  ;on entry
  ;   r0 =   max iterations
  ;   r1 =   p
  ;   r2 =   q
  ;on exit
  ;   r0 =   no. of iterations
          STMFD     r13!,{r1-r8,r14}
          MOV       r8,r0
          MOV       r3,#0               ;x0
          MOV       r4,#0               ;y0
          MOV       r0,#0
loop
           CMP      r3,#5*(1<<SHIFT)
           CMPLT    r4,#5*(1<<SHIFT)
           LDMGEFD  r13!,{r1-r8,pc}
           CMN      r3,#5*(1<<SHIFT)
           CMNGT    r4,#5*(1<<SHIFT)
           LDMLTFD  r13!,{r1-r8,pc}

           MUL      r6,r3,r4
           ADD      r6,r2,r6,ASR#(SHIFT-1)
           MUL      r5,r3,r3
           MUL      r7,r4,r4
           SUB      r5,r5,r7
           ADD      r3,r1,r5,ASR#SHIFT
           MOV      r4,r6
          ADD       r0,r0,#1
          CMP       r0,r8
          BLT       loop
          LDMFD     r13!,{r1-r8,pc}
